# Generated by Django 3.0.3 on 2020-04-01 14:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('avishan', '0009_auto_20200331_2041'),
    ]

    operations = [
        migrations.AddField(
            model_name='activity',
            name='data',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='activity',
            name='object_class',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='activity',
            name='object_id',
            field=models.BigIntegerField(blank=True, default=None, null=True),
        ),
    ]
