# Generated by Django 3.0.3 on 2020-02-15 22:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('avishan', '0004_usergroup_authenticate_with_phone_otp'),
    ]

    operations = [
        migrations.AddField(
            model_name='baseuser',
            name='language',
            field=models.CharField(default='EN', max_length=255),
        ),
    ]
