# Generated by Django 3.0.8 on 2020-07-26 11:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('avishan', '0015_auto_20200724_1419'),
    ]

    operations = [
        migrations.AlterField(
            model_name='country',
            name='flag_url',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
    ]
